"""CSVReader class for extracting data from CSV files."""
import csv
import operator
from datetime import date, datetime
from pathlib import Path

from sc_utility.sc_date_helper import DateHelper


class CSVReader:

    def __init__(self, file_path: Path | str, header_config: list[dict] | None = None):
        """Initialize the CSVReader with the file path.

        Args:
            file_path (Path | str): The path to the CSV file. If the file does not exist, it won't be created.
            header_config (Optional(list[dict]), optional): The header configuration for the CSV file.

        Raises:
            TypeError: If header_config is not structured correctly.
            ImportError: If the file not exists but doesn't have a valid extension.
        """
        # If it's a string, convert to Path
        if isinstance(file_path, str):
            file_path = Path(file_path)

        self.file_exists = False
        self.file_path = file_path
        if header_config is None:
            self.header_config = []
        else:
            self.header_config = header_config
            error_msg = self._validate_header()
            if error_msg:
                raise TypeError(error_msg)

        # Check if the file exists
        if self.file_path.is_file():
            # Check extension to see if it's an Excel file
            if self.file_path.suffix.lower() in {".csv", ".txt"}:
                self.file_exists = True
            else:
                msg = f"File {self.file_path} is not a valid Excel file."
                raise ImportError(msg)

    def _validate_header(self) -> str:  # noqa: PLR0911, PLR0912
        """Validate the header configuration has the allowed structure.

            If header_config has been supplied, validate it:
            1. Make sure it's a list of dictionaries.
            2. Ensure each header is a dictionary with 'name' and 'type' keys.
            3. 'type' can be 'str', 'int', 'float', 'date' or 'datetime'.
            4. If 'date' or 'datetime', it can have an optional 'format' key.
            5. The only allowed attributes in the header dictionaries are 'name', 'type', 'format', 'match' and 'sort'

        Returns:
            str: An error message if the header configuration is invalid, otherwise an empty string.
        """
        if not self.header_config:
            return ""

        if not isinstance(self.header_config, list):
            return "header_config must be a list of dictionaries."

        allowed_keys = {"name", "type", "format", "match", "sort", "minimum"}
        allowed_types = {"str", "int", "float", "date", "datetime"}
        for header in self.header_config:
            if not isinstance(header, dict):
                return "Each header in header_config must be a dictionary."
            # If keys are not one of the allowed keys, return an error
            if not set(header.keys()).issubset(allowed_keys):
                return f"Invalid keys in header configuration. Allowed keys are: {', '.join(allowed_keys)}."
            if "name" not in header or "type" not in header:
                return "Each header must have 'name' and 'type' keys."
            if header["type"] not in allowed_types:
                return f"Invalid type '{header['type']}' in header configuration. Allowed types are: {', '.join(allowed_types)}."

        # If 'format' is provided, it must be a string
        for header in self.header_config:
            if "format" in header and not isinstance(header["format"], str):
                return "If 'format' is provided, it must be a string."

        # If 'sort' is provided, it must be an int
        for header in self.header_config:
            if "sort" in header and not isinstance(header["sort"], int):
                return "If 'sort' is provided, it must be an int."

        # If 'match' is provided, it must be a boolean
        for header in self.header_config:
            if "match" in header and not isinstance(header["match"], bool):
                return "If 'match' is provided, it must be a boolean."

        # If 'minimum' is provided, it must be a date or an int or None
        # and the type must be Date or Datetime
        for header in self.header_config:
            if "minimum" in header and not (isinstance(header["minimum"], (int, date)) or header["minimum"] is None):
                return "If 'minimum' is provided, it must be a date, and int or None."
            if "minimum" in header and header["type"] != "date" and header["type"] != "datetime":
                return "If 'minimum' is provided, the type must be 'date' or 'datetime'."

        return ""

    def read_csv(self) -> list[dict] | None:  # noqa: PLR0912
        """Read the CSV file and return its content.

        If the file does not exist, return None. If the file has a header but no data, returns an empty list.

        Raises:
            ImportError: If the CSV file is empty or has no header.
            ValueError: A value read from the CSV file cannot be converted to the expected type as defined in header_config.

        Returns:
            data (list[dict]): A list of rows from the CSV file or None if the file does not exist.
        """
        # Return None if the file does not exist
        if not self.file_exists:
            return None

        # Read the CSV file
        csv_data = []
        with self.file_path.open(newline="", encoding="utf-8") as csvfile:  # noqa: PLR1702
            reader = csv.reader(csvfile)

            # Read the header
            file_headers = next(reader, None)
            if file_headers is None:
                error_msg = f"CSV file {self.file_path} is empty or has no header."
                raise ImportError(error_msg)

            """ Validate the header against the provided configuration:
            1. All the header names in self.header_config must exist in file_headers, but not necessarily in the same order.
            2. If additional headers are present in the CSV file that are not in self.header_config, add them to file_headers with a type of 'str'.
            """
            header_names = []
            if self.header_config:
                # Check if all headers in header_config are in file_headers
                header_names = [header["name"] for header in self.header_config]
                missing_headers = [
                    header for header in header_names if header not in file_headers
                ]
                if missing_headers:
                    error_msg = f"CSV file {self.file_path} is missing headers: {', '.join(missing_headers)}."
                    raise ImportError(error_msg)

            # Add any additional headers from the CSV file that are not in header_config
            additional_headers = [
                header for header in file_headers if header not in header_names
            ]
            for additional_header in additional_headers:
                self.header_config.append(
                    {"name": additional_header, "type": "str"}
                )

            # Reorder self.header_config to match the headers in the CSV file
            if file_headers is not None:
                self.header_config.sort(key=lambda x: file_headers.index(x["name"]))  # type: ignore[call-arg]

            # Read the data rows from the CSV file
            for row in reader:
                if row:
                    # Convert the row to a dictionary using the self.header_config to convert the data types
                    row_dict = {}
                    for i, header in enumerate(file_headers):
                        if i < len(self.header_config):
                            try:
                                config = self.header_config[i]
                                if config["type"] == "date":
                                    # Convert date strings to datetime objects
                                    date_format = config.get("format", "%Y-%m-%d")
                                    row_dict[header] = DateHelper.parse_date(row[i], date_format)
                                elif config["type"] == "datetime":
                                    # Convert datetime strings to datetime objects
                                    datetime_format = config.get("format", "%Y-%m-%d %H:%M:%S")
                                    # local_tz = datetime.now().astimezone().tzinfo
                                    row_dict[header] = datetime.strptime(row[i], datetime_format)   # .replace(tzinfo=local_tz)  # noqa: DTZ007
                                elif config["type"] == "float":
                                    # Convert float strings to float and round if specified
                                    row_dict[header] = float(row[i])
                                    if "format" in config:
                                        row_dict[header] = float(format(row_dict[header], config["format"]))
                                elif config["type"] == "int":
                                    # Convert Int strings to int
                                    row_dict[header] = int(row[i])
                                else:
                                    # Default to string type
                                    row_dict[header] = row[i]
                            except (ValueError, TypeError) as e:
                                error_msg = (
                                    f"Value '{row[i]}' in row {reader.line_num} cannot be converted to the expected type "
                                    f"for column '{header}'."
                                )
                                raise ValueError(error_msg) from e
                        else:
                            # default to string type for additional headers
                            row_dict[header] = row[i]

                    csv_data.append(row_dict)

            return csv_data

    def sort_csv_data(self, csv_data: list[dict]) -> list[dict]:
        """Sort the CSV data based on the header configuration.

        Args:
            csv_data (list[dict]): The data read from the CSV file.

        Returns:
            list[dict]: The sorted data.
        """
        if not self.header_config:
            return csv_data

        # Get headers that have a 'sort' key and sort them by sort order
        sort_headers = [
            header for header in self.header_config
            if "sort" in header
        ]

        # If no sort headers, return original data
        if not sort_headers:
            return csv_data

        # Sort the headers by their sort value (ascending order)
        sort_headers.sort(key=operator.itemgetter("sort"))

        # Sort the data by each sort header in order
        sorted_data = sorted(
            csv_data,
            key=lambda row: tuple(row[header["name"]] for header in sort_headers)
        )

        return sorted_data

    def merge_data_sets(self, primary_list: list[dict], append_list: list[dict]) -> list[dict]:
        """Merges two lists of dictionaries based on header configuration match fields and sorts the result.

        Args:
            primary_list (list[dict]): The primary list of dictionaries.
            append_list (list[dict]): The list of dictionaries to append or override.

        Raises:
            ValueError: If the dictionaries in the lists do not have the same structure.

        Returns:
            list[dict]: The merged and sorted list of dictionaries.
        """
        # Validate inputs
        if not primary_list and not append_list:
            return []

        def validate_structure(ref_list: list[dict], other_list: list[dict]):
            if not ref_list or not other_list:
                return
            ref_keys = set(ref_list[0].keys())
            for item in other_list:
                if set(item.keys()) != ref_keys:
                    error_msg = "Dictionaries in both lists must have the same structure."
                    raise ValueError(error_msg)

        try:
            validate_structure(primary_list, append_list)
            validate_structure(append_list, primary_list)
        except ValueError as e:
            raise ValueError(e) from e

        # Get match fields from header_config
        match_fields = [
            header["name"] for header in self.header_config
            if header.get("match", False) is True
        ]

        # If no match fields, simply append
        if not match_fields:
            merged = primary_list + append_list
        else:
            def match_key(d):
                return tuple(d[field] for field in match_fields)

            # Build a lookup from primary_list
            primary_map = {match_key(d): d for d in primary_list}

            # Override or add from append_list
            for d in append_list:
                primary_map[match_key(d)] = d

            merged = list(primary_map.values())

        return merged

    def trim_csv_data(self, csv_data: list[dict], max_lines: int | None = None) -> list[dict]:
        """Trim the CSV data based on the header configuration and optionally the max_lines arg.

        Args:
            csv_data (list[dict]): The data read from the CSV file.
            max_lines (Optional(int), optional): If provided, the maximum number of lines to return from csv_data.
                If this is >0 then it will return the first max_lines lines, if <0 then it will return all but the last abs(max_lines) lines. If None, no trimming is done.

        Returns:
            list[dict]: The trimmed data.
        """
        if not self.header_config:
            return csv_data

        # Get headers that have a 'minimum' key
        minimum_headers = [
            header for header in self.header_config
            if "minimum" in header
        ]

        # If no minimum headers and no max_lines, return original data
        if not minimum_headers and max_lines is None:
            return csv_data

        # Trim based on header configuration
        trimmed_data = csv_data
        for header in minimum_headers:
            field_name = header["name"]
            minimum_value = header["minimum"]

            # Calculate the cutoff date
            if isinstance(minimum_value, date):
                cutoff_date = minimum_value
            elif isinstance(minimum_value, int):
                # Calculate date that is minimum_value days before today
                cutoff_date = DateHelper.today_add_days(-minimum_value)
            else:
                continue  # Skip if minimum is neither date nor int

            # Filter out records with dates prior to cutoff_date
            trimmed_data = [
                row for row in trimmed_data
                if field_name in row and isinstance(row[field_name], date) and row[field_name] >= cutoff_date
            ]

        # If max_lines is specified, trim the data to that many lines
        if max_lines is not None and max_lines != 0:
            if max_lines > 0:
                trimmed_data = trimmed_data[:max_lines]
            else:  # max_lines < 0
                trimmed_data = trimmed_data[max_lines:]

        return trimmed_data

    def write_csv(self, data: list[dict], new_filename: Path | str | None = None) -> bool:
        """Write data to the CSV file.

        1. If the file does not exist, it will be created.
        2. If the file exists, it will be overwritten.
        3. The header will be written based on the header_config.
        4. The data will be written in the order of the header_config.
        5. If a header in the data does not exist in the header_config, it will be ignored.
        6. If a header in the header_config does not exist in the data, throw an exception.
        7. Date fields are formatted according to the format specified in header_config.

        Args:
            data (list[dict]): The data to write to the CSV file.
            new_filename (Optional(Path | str), optional)): If provided, the data will be written to this file instead of the original file.

        Raises:
            ValueError: If the data is empty or if a header in header_config is not found in the data.

        Returns:
            True if the data was written successfully, False otherwise.
        """
        if new_filename is not None:
            # If a new filename is provided, use it
            if isinstance(new_filename, str):
                new_filename = Path(new_filename)
            self.file_path = new_filename
            # Check if the file exists
            self.file_exists = bool(self.file_path.is_file())

        if not self.file_exists:
            # Create the file if it does not exist
            self.file_path.parent.mkdir(parents=True, exist_ok=True)
            self.file_path.touch()
            self.file_exists = True

        # Validate the data against the header_config
        if not data:
            error_msg = "No data to write to CSV file."
            raise ValueError(error_msg)

        # Check if all headers in header_config are in the data
        for header in self.header_config:
            if header["name"] not in data[0]:
                error_msg = f"Header '{header['name']}' not found in data."
                raise ValueError(error_msg)

        # Format the data according to header_config
        formatted_data = []
        for row in data:
            formatted_row = {}
            for header in self.header_config:
                field_name = header["name"]
                value = row[field_name]

                # Format according to header config
                if header["type"] == "date" and isinstance(value, date):
                    date_format = header.get("format", "%Y-%m-%d")
                    formatted_row[field_name] = value.strftime(date_format)
                elif header["type"] == "datetime" and isinstance(value, datetime):
                    datetime_format = header.get("format", "%Y-%m-%d %H:%M:%S")
                    formatted_row[field_name] = value.strftime(datetime_format)
                elif header["type"] == "float" and "format" in header:
                    formatted_row[field_name] = format(value, header["format"])
                else:
                    formatted_row[field_name] = value

            formatted_data.append(formatted_row)

        # Write the CSV file
        with self.file_path.open("w", newline="", encoding="utf-8") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=[header["name"] for header in self.header_config])
            writer.writeheader()
            writer.writerows(formatted_data)

        return True

    def update_csv_file(self, new_data: list[dict], new_filename: Path | str | None = None) -> list[dict]:
        """Appends or merges the new_data into an existing CSV file. If the file does not exist, it will be created.

        This function will also sort and trim the combined data according to the header configuration.

        Args:
            new_data (list[dict]): The new data to append or merge.
            new_filename (Optional(Path | str), optional)): If provided, the data will be written to this file instead of the original file.

        Raises:
            RuntimeError: If there is a problem processign the data.

        Returns:
            merged_data (list[dict]): The merged and sorted data after appending or merging the new_data.
        """
        # Read the existing CSV data
        try:
            current_data = self.read_csv()
        except (ImportError, TypeError, ValueError) as e:
            raise RuntimeError(e) from e

        if current_data is not None:
            # Merge / append the extra data
            try:
                merged_data = self.merge_data_sets(current_data, new_data)
            except ValueError as e:
                raise RuntimeError(e) from e

            # Sort the data according to the header configuration
            merged_data = self.sort_csv_data(merged_data)

            # Trim the CSV data
            merged_data = self.trim_csv_data(merged_data)

        # Save the modified CSV data
        self.write_csv(merged_data, new_filename)
        return merged_data



"""
Spello Consulting Logging Module.

Provides general purpose logging functions.
"""

import inspect
import os
import platform
import smtplib
import sys
import traceback
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path


class SCLogger:
    """A class to handle logging messages with different verbosity levels."""

    def __init__(self, logger_settings: dict | None = None, logfile_name: str | None = None, file_verbosity: str | None = "detailed", console_verbosity: str | None = "summary", max_lines: int | None = 10000):
        """
        Initializes the logger with configuration settings.

        If logger_settings are provided, it will override the individual parameters.

        Args:
            logger_settings (Optional[dict], optional): A dictionary containing logger settings. If provided, it should include the same keys as the individual parameters below:
            logfile_name (Optional[str], optional): The name of the log file. If None, no file logging will be done.
            file_verbosity (Optional[str], optional): Verbosity level for file logging
            console_verbosity  (Optional[str], optional): Verbosity level for console logging
            max_lines  (Optional[int], optional): Maximum number of lines to keep in the log file
        """
        if logger_settings is not None:
            self.logfile_name = logger_settings["logfile_name"]
            self.file_verbosity = logger_settings["file_verbosity"]
            self.console_verbosity = logger_settings["console_verbosity"]
            self.max_lines = logger_settings["max_lines"]
            self.log_process_id = logger_settings["log_process_id"]
        else:
            self.logfile_name = logfile_name
            self.file_verbosity = file_verbosity
            self.console_verbosity = console_verbosity
            self.max_lines = max_lines
            self.log_process_id = False

        self.verbosity_levels = {
            "none": 0,
            "error": 1,
            "warning": 2,
            "summary": 3,
            "detailed": 4,
            "debug": 5,
            "all": 6,
        }

        # Platform specific settings
        self.os_name = self.get_os()

        # Use the register_email_settings method to set up email settings
        self.email_settings = None

        # See if logfile writing is required
        self.file_logging_enabled = self.logfile_name is not None

        # Make a note of the app directory
        self.app_dir = self.client_dir = Path(sys.argv[0]).parent.resolve()

        if self.file_logging_enabled and self.logfile_name is not None:
            # Determine the file path for the log file
            current_dir = Path.cwd()

            self.logfile_path = current_dir / self.logfile_name
            if not self.logfile_path.exists():
                self.logfile_path = self.app_dir / self.logfile_name

            # Truncate the log file if it exists
            self._initialise_monitoring_logfile()

        # Setup the path to the fatal error tracking file
        self.fatal_error_file_path = self.app_dir / f"{self.app_dir.name}_fatal_error.txt"

        # Save process ID
        self.process_id = os.getpid()

    def get_os(self) -> str:  # noqa: PLR6301
        """Returns the name of the operating system.

        Returns:
            os_string (str): The name of the operating system in lowercase.
        """
        # Get the platform name and convert it to lowercase
        platform_name = platform.system().lower()

        if platform_name == "darwin":
            platform_name = "macos"

        return platform_name

    def _initialise_monitoring_logfile(self) -> None:
        """Initialise the monitoring log file. If it exists, truncate it to the max number of lines."""
        if not self.file_logging_enabled:
            return

        self.trim_logfile()

    def trim_logfile(self) -> None:
        """Trims the log file to the maximum number of lines specified."""
        if not self.file_logging_enabled:
            return

        if self.logfile_path.exists() and self.max_lines is not None and self.max_lines > 0:
            # Monitoring log file exists - truncate excess lines if needed.
            with self.logfile_path.open(encoding="utf-8") as file:
                lines = file.readlines()

                if len(lines) > self.max_lines:
                    # Keep the last max_lines rows
                    keep_lines = lines[-self.max_lines:] if len(lines) > self.max_lines else lines

                    # Overwrite the file with only the last 1000 lines
                    with self.logfile_path.open("w", encoding="utf-8") as file2:
                        file2.writelines(keep_lines)

    def log_message(self, message: str, verbosity: str = "summary") -> None:
        """Writes a log message to the console and/or a file based on verbosity settings.

        Args:
            message (str): The message to log.
            verbosity (str): The verbosity level for the message. Must be one of "none", "error", "warning", "summary", "detailed", "debug", or "all".

        Raises:
            ValueError: If an invalid verbosity level is provided.

        """
        if verbosity not in self.verbosity_levels:
            exception_msg = f"log_message(): Invalid verbosity passed, must be one of {list(self.verbosity_levels.keys())}."
            raise ValueError(exception_msg)

        file_verbosity = self.file_verbosity if self.file_verbosity is not None else "none"
        console_verbosity = self.console_verbosity if self.console_verbosity is not None else "none"
        logfile_level = self.verbosity_levels.get(file_verbosity, 0)
        console_level = self.verbosity_levels.get(console_verbosity, 0)
        message_level = self.verbosity_levels.get(verbosity, 0)

        process_str = ""
        if self.log_process_id:
            process_str = f" Proc {self.process_id}"

        # Deal with console message first
        if console_level >= message_level and console_level > 0:
            if verbosity == "error":
                print("ERROR: " + message, file=sys.stderr)
            elif verbosity == "warning":
                print("WARNING: " + message)
            else:
                print(message)

        # Now write to the log file if needed
        if self.file_logging_enabled:
            error_str = " ERROR" if verbosity == "error" else " WARNING" if verbosity == "warning" else ""
            if logfile_level >= message_level and logfile_level > 0:
                with self.logfile_path.open("a", encoding="utf-8") as file:
                    if not message:
                        file.write("\n")
                    else:
                        # Use the local timezone for the log timestamp
                        local_tz = datetime.now().astimezone().tzinfo
                        file.write(f"{datetime.now(local_tz).strftime('%Y-%m-%d %H:%M:%S')}{process_str}{error_str}: {message}\n")

    def register_email_settings(self, email_settings: dict | None) -> None:
        """
        Registers email settings for sending emails.

        Use the SCConfigManager.get_email_settings() method to get a dictionary containing the email settings. Otherwise, you can pass a
        dictionary directly to this method with these keys:
            SendEmailsTo: str, the email address to send emails to
            SMTPServer: str, the SMTP server address
            SMTPPort: int, the SMTP server port (optional, defaults to 587)
            SMTPUsername: str, the username for the SMTP server
            SMTPPassword: str, the password for the SMTP server (preferably an App Password)
            SubjectPrefix: str, a prefix for email subjects (optional, default to None)

        Args:
            email_settings (Optional[dict], optional): A dictionary containing email settings. If empty or None, no email settings will be registered.

        Raises:
            TypeError: If email_settings is not a dictionary.
            ValueError: If any required keys are missing from the email_settings dictionary.

        """
        if email_settings == {} or email_settings is None:
            return  # No email settings provided, so skip registration

        if not isinstance(email_settings, dict):
            msg = "register_email_settings(): Email settings must be a dictionary."
            raise TypeError(msg)

        # Check if all required keys are present
        required_keys = ["SendEmailsTo", "SMTPServer", "SMTPUsername", "SMTPPassword"]
        for key in required_keys:
            if key not in email_settings:
                msg = f"register_email_settings(): Missing required email setting: {key}"
                raise ValueError(msg)

        # Store the email settings in the config object
        self.email_settings = email_settings

    def is_probable_path(self, possible_path: str | Path) -> bool:
        """
        Checks if the given string or Path object is likely to be a file path.

        This method checks if the string is an absolute path, contains a path separator, or has a file extension.

        Args:
            possible_path (str): The string to check.

        Returns:
            result (bool): True if the string is likely a file path, False otherwise.
        """
        max_path = 260 if self.os_name == "windows" else os.pathconf("/", "PC_PATH_MAX")

        path_obj = None
        if isinstance(possible_path, Path):
            path_str = str(possible_path)
            path_obj = possible_path
        else:
            path_str = possible_path

        if len(path_str) > max_path:
            # If the path is longer than the maximum allowed path length, it cannot be a valid path
            return False

        if path_obj is None:
            path_obj = Path(possible_path)

        # Check if it's absolute, or contains a path separator, or has a file extension
        if path_obj.is_absolute():
            return True

        if "/" in path_str or "\\" in path_str:
            return True

        # Check if the path has a file extension
        return bool(path_obj.suffix and path_obj.suffix.lower() is not None)

    def select_file_location(self, file_name: str | Path) -> Path | None:
        """
        Selects the file location for the given file name.

        Args:
            file_name (str): The name of the file to locate. Can be just a file name, or a relative or absolute path.

        Returns:
            location (Path): The full path to the file as a Path object. If the file does not exist in the current directory, it will look in the script directory.
        """
        # Look at the file_name and see if it looks like a path
        if not self.is_probable_path(file_name):
            return None

        # Check to see if file_name is a full path or just a file name
        file_path = Path(file_name)

        # Check if file_name is an absolute path, return this even if it does not exist
        if file_path.is_absolute():
            return file_path

        # Check if file_name contains any parent directories (i.e., is a relative path)
        # If so, return this even if it does not exist
        if file_path.parent != Path("."):  # noqa: PTH201
            # It's a relative path
            return (Path.cwd() / file_path).resolve()

        # Otherwise, assume it's just a file name and look for it in the current directory and the script directory
        current_dir = Path.cwd()
        app_dir = self.client_dir = Path(sys.argv[0]).parent.resolve()
        file_path = current_dir / file_name
        if not file_path.exists():
            file_path = app_dir / file_name
        return file_path

    def send_email(self, subject: str, body: str | Path, test_mode: bool = False) -> bool:  # noqa: FBT001, FBT002, PLR0912, PLR0915
        """
        Sends an email using the SMTP server previously specified in register_email_settings().

        Args:
            subject (str): The subject of the email.
            body (str): The body of the email. This argument can be one of 4 things:
                1. A string containing the HTML body of the email
                2. A string or Path containing the path to an HTML file to read the body from
                3. A string containing the text body of the email
                4. A string or Path containing the path to an text file to read the body from
            test_mode (bool, optional): If True, the email will not be sent, but a message will be logged indicating that the email would have been sent. Defaults to False.

        Returns:
            result (bool): True if the email was sent successfully, False otherwise.
        """
        if self.email_settings is None:
            return False  # No email settings registered, so skip sending the email

        # First confirm that the body argument was passed as a string
        if not isinstance(body, str) and not isinstance(body, Path):
            self.log_fatal_error("body argument must be a string containing the body content or a file path.")
            return False

        # See if the body string is something that looks like a file path
        try:
            payload_path = self.select_file_location(body)
        except OSError:
            # Nothing to do here
            payload_path = None

        payload = body  # Default to the body as text content
        payload_type = "plain"

        # Default to treating the body a file path and see if we can resolve it
        if payload_path is not None:
            if payload_path.exists():
                # If the body is a file path, read the content
                with payload_path.open("r", encoding="utf-8") as file:
                    payload = file.read()

                # Determine the payload type based on the file extension
                if payload_path.suffix.lower() == ".html":
                    payload_type = "html"
                elif payload_path.suffix.lower() == ".txt":
                    payload_type = "plain"
                else:
                    self.log_fatal_error(f"Unsupported file type for email body: {payload_path.suffix}")
                    return False
            else:
                payload_path = None

        if payload_path is None:
            # If fall through to here, then the body was not a file path or the file did not exist
            # Assume that the body is a string containing the content
            payload = str(body)  # Default to the body as text content
            if payload.startswith(("<html", "<!DOCTYPE html")):
                payload_type = "html"
            else:
                payload_type = "plain"

        # Load the Gmail SMTP server configuration
        sender_email = self.email_settings.get("SMTPUsername")
        send_to = self.email_settings.get("SendEmailsTo")
        smtp_server = self.email_settings.get("SMTPServer")
        smtp_port = self.email_settings.get("SMTPPort", 587)
        smtp_password = self.email_settings.get("SMTPPassword")

        if not sender_email or not send_to or not smtp_server or not smtp_port or not smtp_password:
            self.log_fatal_error("send_email(): Email settings are incomplete. 'SMTPUsername', 'SendEmailsTo', 'SMTPServer' and 'SMTPPassword' must be provided.")
            return False

        try:
            # Create the email
            msg = MIMEMultipart()
            msg["From"] = sender_email
            msg["To"] = send_to
            if self.email_settings.get("SubjectPrefix", None) is not None:
                msg["Subject"] = self.email_settings.get("SubjectPrefix", "") + subject
            else:
                msg["Subject"] = subject

            # Attach the payload as either plain text or HTML
            part = MIMEText(payload, payload_type)  # type: ignore[arg-type]
            msg.attach(part)

            # Connect to the Gmail SMTP server
            with smtplib.SMTP(smtp_server, smtp_port) as server:
                server.starttls()  # Upgrade the connection to secure
                server.login(sender_email, smtp_password)  # Log in using App Password
                if test_mode:
                    self.log_message(f"Test mode: Email with subject '{msg['Subject']}' would be sent to {send_to}.", "summary")
                else:
                    server.sendmail(sender_email, send_to, msg.as_string())  # Send the email

        except RuntimeError as e:
            self.log_fatal_error(f"send_email(): Failed to send email with subject {msg['Subject']}: {e}")
            return False

        else:
            return True  # Email sent successfully

    def log_fatal_error(self, message: str, report_stack: bool = False, calling_function: str | None = None) -> None:  # noqa: FBT001, FBT002
        """
        Log a fatal error, send an email if configured to so and then exit the program.

        Args:
            message (str): The error message to log.
            report_stack (Optional[bool], optional): If True, include the stack trace in the log message.
            calling_function (Optional[str], optional): The name of the function that called this method, if known. If None, the calling function will be determined automatically.

        Raises:
            SystemExit: Exits the program with a status code of 1 after logging the fatal error.

        """  # noqa: DOC502
        function_name = None
        if calling_function is None:
            stack = inspect.stack()
            # Get the frame of the calling function
            calling_frame = stack[1]
            # Get the function name
            function_name = calling_frame.function
            if function_name == "<module>":
                function_name = "main"
            # Get the class name (if it exists)
            class_name = None
            if "self" in calling_frame.frame.f_locals:
                class_name = calling_frame.frame.f_locals["self"].__class__.__name__
                full_reference = f"{class_name}.{function_name}()"
            else:
                full_reference = function_name + "()"
        else:
            full_reference = calling_function + "()"

        if report_stack:
            stack_trace = traceback.format_exc()
            message += f"\n\nStack trace:\n{stack_trace}"

        self.log_message(f"Function {full_reference}: FATAL ERROR: {message}", "error")

        # Try to send an email if configured to do so and we haven't already sent one for a fatal error
        # Don't send concurrent error emails
        if function_name != "send_email" and not self.get_fatal_error():
            self.send_email(
                f"{self.app_dir.name} terminated with a fatal error",
                f"{message} \nAdditional emails will not be sent for concurrent errors.",
            )

        # record the error in in a file so that we keep track of this next time round
        self.set_fatal_error(message)

        # Exit the program
        sys.exit(1)

    def get_fatal_error(self) -> bool:
        """Returns True if a fatal error was previously reported, false otherwise."""
        return Path(self.fatal_error_file_path).exists()

    def clear_fatal_error(self) -> bool:
        """
        Clear a previously logged fatal error.

        Returns:
            result (bool): True if the file was deleted, False if it did not exist.
        """
        if Path(self.fatal_error_file_path).exists():
            Path(self.fatal_error_file_path).unlink()
            return True
        return False

    def set_fatal_error(self, message: str) -> None:
        """
        Create a fatal error tracking file and write the message to it.

        Args:
            message (str): The error message to write to the fatal error file.
        """
        with Path(self.fatal_error_file_path).open("w", encoding="utf-8") as file:
            file.write(message)

    def get_process_id(self) -> int:
        """
        Returns the process ID of the current process.

        Returns:
            process_id (int): The process ID of the current process.
        """
        return self.process_id



"""Spello Consulting Configuration Manager Module.

Management of a YAML log file.
"""

import sys
from collections.abc import Callable
from pathlib import Path

import yaml
from cerberus import Validator


class SCConfigManager:
    """Loads the configuration from a YAML file, validates it, and provides access to the configuration values."""

    def __init__(self, config_file: str, default_config: dict | None = None, validation_schema: dict | None = None, placeholders: dict | None = None):
        """Initializes the configuration manager.

        Args:
            config_file (str): The relative or absolute path to the configuration file.
            default_config (Optional[dict], optional): A default configuration dict to use if the config file does not exist.
            validation_schema (Optional[dict], optional): A cerberus style validation schema dict to validate the config file against.
            placeholders (Optional[dict], optional): A dictionary of placeholders to check in the config. If any of these are found, a exception will be raised.

        Raises:
            RuntimeError: If the config file does not exist and no default config is provided, or if there are YAML errors in the config file.

        """
        self._config = {}    # Intialise the actual config object
        self.config_file = config_file
        self.config_last_modified = None  # Last modified time of the config file
        self.logger_function = None  # Placeholder for a logger function
        self.validation_schema = validation_schema
        self.placeholders = placeholders

        # Make a note of the app directory
        self.app_dir = self.client_dir = Path(sys.argv[0]).parent.resolve()

        # Determine the file path for the log file
        self.config_path = self.select_file_location(self.config_file)

        # If the config file doesn't exist and we have a default config, write that to file
        if not self.config_path.exists():
            if default_config is None:
                msg = f"Cannot find config file {self.config_file} and no default config provided."
                raise RuntimeError(msg)

            with Path(self.config_path).open("w", encoding="utf-8") as file:
                yaml.dump(default_config, file)

        # Now load the config file
        self.load_config()

    def load_config(self) -> bool:
        """Load the configuration from the config file specified to the __init__ method.

        Raises:
            RuntimeError: If there are YAML errors in the config file, if placeholders are found, or if validation fails.

        Returns:
            result (bool): True if the configuration was loaded successfully, otherwise False.
        """
        with Path(self.config_path).open(encoding="utf-8") as file:
            try:
                self._config = yaml.safe_load(file)

            except yaml.YAMLError as e:
                msg = f"YAML error in config file {self.config_file}: {e}"
                raise RuntimeError(msg) from e

            else:
                # Make sure there are no placeholders in the config file, exit if there are
                self.check_for_placeholders(self.placeholders)

                # If we have a validation schema, validate the config
                if self.validation_schema is not None:
                    v = Validator()

                    if not v.validate(self._config, self.validation_schema):  # type: ignore[call-arg]
                        msg = f"Validation error for config file {self.config_path}: {v.errors}"  # type: ignore[call-arg]
                        raise RuntimeError(msg)

        self.config_last_modified = self.config_path.stat().st_mtime
        return True

    def check_for_config_changes(self) -> bool:
        """Check if the configuration file has changed. If it has, reload the configuration.

        Returns:
            result (bool): True if the configuration has changed, False otherwise.
        """
        # get the last modified time of the config file
        last_modified = self.config_path.stat().st_mtime

        if self.config_last_modified is None or last_modified > self.config_last_modified:
            # The config file has changed, reload it
            self.load_config()
            self.config_last_modified = last_modified
            return True

        return False

    def select_file_location(self, file_name: str) -> Path:
        """Selects the file location for the given file name.

        Args:
            file_name (str): The name of the file to locate. Can be just a file name, or a relative or absolute path.

        Returns:
            file_path (Path): The full path to the file as a Path object. If the file does not exist in the current directory, it will look in the script directory.
        """
        # Check to see if file_name is a full path or just a file name
        file_path = Path(file_name)

        # Check if file_name is an absolute path, return this even if it does not exist
        if file_path.is_absolute():
            return file_path

        # Check if file_name contains any parent directories (i.e., is a relative path)
        # If so, return this even if it does not exist
        if file_path.parent != Path("."):  # noqa: PTH201
            # It's a relative path
            return (Path.cwd() / file_path).resolve()

        # Otherwise, assume it's just a file name and look for it in the current directory and the script directory
        current_dir = Path.cwd()
        app_dir = self.client_dir = Path(sys.argv[0]).parent.resolve()
        file_path = current_dir / file_name
        if not file_path.exists():
            file_path = app_dir / file_name
        return file_path

    def register_logger(self, logger_function: Callable) -> None:
        """Registers a logger function to be used for logging messages.

        Args:
            logger_function (Callable): The function to use for logging messages.
        """
        self.logger_function = logger_function

    def check_for_placeholders(self, placeholders: dict | None) -> bool:
        """Recursively scan self._config for any instances of a key found in placeholders.

        If the keys and values match (including nested), return True.

        Args:
            placeholders (dict): A dictionary of placeholders to check in the config.

        Raises:
            RuntimeError: If any placeholder is found in the config file, an exception will be raised with a message indicating the placeholder and its value.

        Returns:
            result (bool): True if any placeholders are found in the config, otherwise False.
        """  # noqa: DOC502
        def recursive_check(config_section, placeholder_section):
            for key, placeholder_value in placeholder_section.items():
                if key in config_section:
                    config_value = config_section[key]
                    if isinstance(placeholder_value, dict) and isinstance(config_value, dict):
                        if recursive_check(config_value, placeholder_value):
                            return True
                    elif config_value == placeholder_value:
                        msg = f"Placeholder value '{key}: {placeholder_value}' found in config file {self.config_path}. Please fix this."
                        raise RuntimeError(msg)
            return False

        if placeholders is None:
            return False

        return recursive_check(self._config, placeholders)

    def get(self, *keys, default=None):
        """Retrieve a value from the config dictionary using a sequence of nested keys.

        Example:
            value = config_mgr.get("DeviceType", "WebsiteAccessKey")

        Args:
            keys (*keys): Sequence of keys to traverse the config dictionary.
            default (Optional[variable], optional): Value to return if the key path does not exist.

        Returns:
            value (variable): The value if found, otherwise the default.

        """
        value = self._config
        try:
            for key in keys:
                value = value[key]
        except (KeyError, TypeError):
            return default
        else:
            return value

    def get_logger_settings(self, config_section: str | None = "Files") -> dict:
        """Returns the logger settings from the config file.

        Args:
            config_section (Optional[str], optional): The section in the config file where logger settings are stored.

        Returns:
            settings (dict): A dictionary of logger settings that can be passed to the SCLogger() class initialization.
        """
        logger_settings = {
            "logfile_name": self.get(config_section, "LogfileName", default="default_logfile.log"),
            "file_verbosity": self.get(config_section, "LogfileVerbosity", default="detailed"),
            "console_verbosity": self.get(config_section, "ConsoleVerbosity", default="summary"),
            "max_lines": self.get(config_section, "LogfileMaxLines", default=10000),
            "log_process_id": self.get(config_section, "LogProcessId", default=False),
        }
        return logger_settings

    def get_email_settings(self, config_section: str | None = "Email") -> dict | None:
        """Returns the email settings from the config file.

        Args:
            config_section (Optional[str], optional): The section in the config file where email settings are stored.

        Returns:
            settings (dict): A dictionary of email settings or None if email is disabled or not configured correctly.
        """
        # fir check to see if we have an EnableEmail setting
        enable_email = self.get(config_section, "EnableEmail", default=True)
        if not enable_email:
            return None

        email_settings = {
            "SendEmailsTo": self.get(config_section, "SendEmailsTo"),
            "SMTPServer": self.get(config_section, "SMTPServer"),
            "SMTPUsername": self.get(config_section, "SMTPUsername"),
            "SMTPPassword": self.get(config_section, "SMTPPassword"),
            "SubjectPrefix": self.get(config_section, "SubjectPrefix"),
        }

        # Only return true if all the required email settings have been specified (excluding SubjectPrefix)
        required_fields = {k: v for k, v in email_settings.items() if k != "SubjectPrefix"}
        if all(required_fields.values()):
            return email_settings

        return None



"""
sc_utility package.

This package provides utility functions and classes for the SC project.
"""
from .sc_config_mgr import SCConfigManager
from .sc_csv_reader import CSVReader
from .sc_date_helper import DateHelper
from .sc_excel_reader import ExcelReader
from .sc_logging import SCLogger

__all__ = ["CSVReader", "DateHelper", "ExcelReader", "SCConfigManager", "SCLogger"]



"""ExcelReader class for extracting data from Excel files."""
import warnings
import zipfile
from pathlib import Path

import pandas as pd
from openpyxl import load_workbook
from openpyxl.utils import get_column_letter, range_boundaries
from openpyxl.utils.exceptions import InvalidFileException


class ExcelReader:
    """
    A class to read and extract data from Excel files (.xlsx, .xlsm, .xlsb).

    This class provides methods to load workbooks, extract data from sheets,
    tables, and named ranges, with robust error handling.
    """

    def __init__(self, file_path: Path | str):
        """
        Initializes the ExcelReader with the path to the Excel file.

        Args:
            file_path (str | Path): Path to the Excel file, specified as a Path object or string.

        Raises:
            ImportError: If the file does not exist, is not a valid Excel file, or if the openpyxl library cannot be imported.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)
        if not file_path.is_file():
            msg = f"File {file_path} does not exist or is not a valid file."
            raise ImportError(msg)

        # Check extension to see if it's an Excel file
        if file_path.suffix.lower() not in {".xlsx", ".xlsm", ".xlsb"}:
            msg = f"File {file_path} is not a valid Excel file."
            raise ImportError(msg)

        self.file_path = file_path

    def load_excel_workbook(self, data_only=True, read_only=False):  # noqa: FBT002
        """
        Load an Excel workbook with robust error handling.

        Args:
            data_only (Optional[bool], optional): Whether to return cell values (not formulas).
            read_only (Optional[bool], optional): Use openpyxl's read-only mode for large files.

        Raises:
            ImportError: If the file cannot be loaded due to various reasons (e.g., file not found, permission denied, invalid format).

        Returns:
            object (workbook): Workbook object or None if loading fails.
        """
        warnings.filterwarnings("ignore", category=UserWarning, module="openpyxl")
        try:
            wb = load_workbook(filename=self.file_path, data_only=data_only, read_only=read_only)
        except FileNotFoundError:
            msg = f"File not found: {self.file_path}"
        except PermissionError:
            msg = f"Permission denied: {self.file_path}"
        except InvalidFileException:
            msg = f"Invalid Excel file format: {self.file_path}"
        except zipfile.BadZipFile:
            msg = f"Corrupt Excel file (not a valid ZIP): {self.file_path}"
        except OSError as e:
            msg = f"Unexpected error loading Excel file {self.file_path}: {e}"
        else:
            if wb is None:
                msg = f"Failed to load workbook from {self.file_path}. The file may be corrupted or not a valid Excel file."
            else:
                return wb

        raise ImportError(msg)   # Raise an error with the message if loading fails

    def extract_data(self, source_name: str, source_type: str) -> list[dict]:
        """
        Extracts data from an Excel file based on the source type and name.

        Expected the specified source type to be either:
        - An entire worksheet with the header in the first row (sheet)
        - A named range (range)
        - An Excel table (table)

        Args:
            source_name (str): Name of the sheet, table, or range to extract.
            source_type (str): Type of source ('sheet', 'table', or 'range').

        Raises:
            ImportError: If the source type is invalid or if there are issues extracting data.

        Returns:
            data (list[dict]): Data extracted as a dictionary.
        """
        if source_type == "sheet":
            return self.extract_from_sheet(source_name)
        if source_type == "table":
            return self.extract_from_table(source_name)
        if source_type == "range":
            return self.extract_from_range(source_name)

        msg = f"Invalid source type '{source_type}'. Must be 'sheet', 'table', or 'range'."
        raise ImportError(msg)

    def extract_from_sheet(self, sheet_name: str) -> list[dict]:
        """
        Extracts a sheet from an Excel file and returns it as a DataFrame.

        Args:
            sheet_name (str): Name of the sheet to extract.

        Raises:
            ImportError: If the sheet cannot be loaded or if there are issues with the file.

        Returns:
            data (list[dict]): A list containing the sheet data.
        """
        try:
            table_df = pd.read_excel(self.file_path, sheet_name=sheet_name)
            table_df.columns = table_df.columns.str.strip()
            table_data = table_df.to_dict(orient="records")
        except ImportError as e:
            msg = f"Error importing the openpyxl library for reading Excel file {self.file_path}: {e}. "
            msg += "Try installing it with 'pip install openpyxl' or 'uv add openpyxl'."
        except zipfile.BadZipFile:
            msg = f"Corrupt Excel file (not a valid ZIP): {self.file_path}"
        except OSError as e:
            msg = f"Unexpected error loading Excel file {self.file_path}: {e}"
        except (ValueError, AttributeError) as e:
            msg = f"Error loading sheet '{sheet_name}' from Excel file {self.file_path}: {e}"
        else:
            return table_data

        # If we reach here, it means the import failed
        raise ImportError(msg)

    def extract_from_table(self, table_name: str) -> list[dict]:
        """
        Extracts a table from an Excel file and returns it as a DataFrame.

        Args:
            table_name (str): Name of the table to extract.

        Raises:
            ImportError: If the table cannot be loaded or if there are issues with the file.

        Returns:
            data (list[dict]): A list containing the table data.
        """
        # load_excel_workbook() will raise an ImportError if there's an issue. Let this be caught by the caller.
        wb = self.load_excel_workbook(data_only=True, read_only=False)
        found = False
        selected_sheet = None
        for sheet_name in wb.sheetnames:
            ws = wb[sheet_name]
            if table_name in ws.tables:
                table = ws.tables[table_name]
                table_range = table.ref  # e.g., "B3:F12"
                min_col, min_row, max_col, max_row = range_boundaries(table_range)
                found = True
                selected_sheet = sheet_name
                break  # Stop scanning after finding the first match

        if not found:
            msg = f"Table '{table_name}' not found in any worksheet of the Excel file {self.file_path}."
            raise ImportError(msg)

        if min_col is None or max_col is None or max_row is None or min_row is None:
            msg = f"Invalid table range boundaries for table '{table_name}' in file {self.file_path}."
            raise ImportError(msg)

        try:
            # Generate usecols as Excel letters (e.g., 'B', 'C', ..., 'F')
            usecols = [get_column_letter(col) for col in range(min_col, max_col + 1)]
            # Include all rows in the table
            nrows = max_row - min_row + 1  # type: ignore[assignment]

            # Read the specific table range using pandas
            table_df = pd.read_excel(
                self.file_path,
                sheet_name=selected_sheet,
                usecols=",".join(usecols),
                skiprows=min_row - 1,   # skip up to the header row (Excel is 1-based)
                nrows=nrows,
                header=None     # Don't treat the first row as header so that we avoid the bug where pandas renames columns it thinks are duplicates
            )

            # Strip whitespace from column names
            table_df.columns = table_df.iloc[0].map(str).str.strip()  # type: ignore[attr-defined]

            # Define the first row as the header
            table_df = table_df[1:].reset_index(drop=True)  # type: ignore[attr-defined]

            # Extract to dictionary
            table_data = table_df.to_dict(orient="records")

        except (ImportError, ValueError, AttributeError) as e:
            msg = f"Error loading table '{table_name}' from Excel file {self.file_path}: {e}"
            raise ImportError(msg) from e
        else:
            return table_data

    def extract_from_range(self, range_name: str) -> list[dict]:
        """
        Extracts a table from an Excel file and returns it as a DataFrame.

        Args:
            range_name (str): Name of the range to extract.

        Raises:
            ImportError: If the table cannot be loaded or if there are issues with the file.

        Returns:
            data (list[dict]): A list containing the range data.
        """
        # load_excel_workbook() will raise an ImportError if there's an issue. Let this be caught by the caller.
        wb = self.load_excel_workbook(data_only=True, read_only=True)
        found = False

        for defined_name in wb.defined_names:
            if defined_name == range_name:
                dn = wb.defined_names[defined_name]
                destinations = list(dn.destinations)
                if len(destinations) != 1:
                    msg = f"Range '{range_name}' refers to multiple areas, which is unsupported."
                    raise ImportError(msg)

                sheet_name, ref = destinations[0]  # unpack tuple
                min_col, min_row, max_col, max_row = range_boundaries(ref)
                found = True
                break

        if not found:
            msg = f"Range '{range_name}' not found in any worksheet of the Excel file {self.file_path}."
            raise ImportError(msg)

        if min_col is None or max_col is None or max_row is None or min_row is None:
            msg = f"Invalid table range boundaries for range '{range_name}' in file {self.file_path}."
            raise ImportError(msg)

        try:
            # Generate usecols as Excel letters (e.g., 'B', 'C', ..., 'F')
            usecols = [get_column_letter(col) for col in range(min_col, max_col + 1)]
            # Include all rows in the table
            nrows = max_row - min_row + 1    # type: ignore[assignment]

            # Read the specific table range using pandas
            table_df = pd.read_excel(
                self.file_path,
                sheet_name=sheet_name,
                usecols=",".join(usecols),
                skiprows=min_row - 1,  # skip to the first row of the range
                nrows=nrows,
                header=None     # Don't treat the first row as header so that we avoid the bug where pandas renames columns it thinks are duplicates
            )

            # Strip whitespace from column names
            table_df.columns = table_df.iloc[0].map(str).str.strip()  # type: ignore[attr-defined]

            # Define the first row as the header
            table_df = table_df[1:].reset_index(drop=True)

            # Extract to dictionary
            table_data = table_df.to_dict(orient="records")

        except (ImportError, ValueError, AttributeError) as e:
            msg = f"Error loading data range '{range_name}' from Excel file {self.file_path}: {e}"
            raise ImportError(msg) from e
        else:
            return table_data


